(* Global types, constants and subroutines for
   ExtFileTools, AesLib, FtpFileTools

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1.0 - Jun. 2006
   vers. 2.0 - Apr. 2007 : Error messages changed
   *)

unit FtGlobal;

interface

type
  TFileAction = (acNone,acCopy,acCompress,acUnCompress,acEncrypt,acDecrypt,
                 acVerify,acAddVerify,acFtpWrite,acFtpRead);
  TProgressEvent = procedure(AAction : TFileAction; ACount: int64) of object;

const
  // error codes
  errOK         = 0;
  errFileCreate = 1;
  errFileOpen   = 2;
  errFileClose  = 3;
  errFileRead   = 4;
  errFileWrite  = 5;
  errFileAttr   = 6;  // Error setting attr.
  errFileFull   = 7;
  errFileGZip   = 8;  // ill. GZip header
  errFileCheck  = 9;  // corrupt copied or packed file
  errFileECrypt = 10; // error encrypting file
  errFileDCrypt = 11; // error decrypting file
  errFileVerify = 12; // verify error
  errLongPath   = 13; // path too long (>260)
  errNotFound   = 14; // source file not found
  errFileTS     = 15; // Error setting timestamp after copy
  errStorage    = 16; // Error copying file details
  errFtpRead    = 24; // error reading via FTP
  errFtpWrite   = 25; // error writing via FTP
  errFtpConnect = 26; // connection error on FTP
  errFtpBroken  = 27; // connection reset by server
  errFTPDatConn = 28; // error opening data connection
  errSignature  = 32; // ill. Zip signature
  errExtract    = 33; // Error extracting file
  errFormat     = 34; // Unsupported file format

  errUserBreak  = $80; // process stopped bei user
  errAllCodes   = $00FF;

  errError      = $0100;
  errWarning    = $0200;
  errAllTypes   = $0F00;

  errCopy       = $1000;
  errGZip       = $2000;
  errGUnzip     = $3000;
  errZip        = $4000;
  errEncrypt    = $5000;
  errDecrypt    = $6000;
  errAllSources = $F000;
  
function rsErrOpening : string;
function rsErrCreating : string;
function rsErrReading : string;
function rsErrWriting : string;
function rsErrClosing : string;
function rsErrSetAttr : string;
function rsErrTimeStamp : string;
function rsErrDirEntry : string;
function rsErrEndRecord : string;
function rsErrVerify : string;
function rsErrCopy : string;

// get error message
function GetCopyErrMsg (AError : integer) : string;

implementation

uses SysUtils, GnuGetText;

{ ------------------------------------------------------------------- }
// get error message
function rsErrOpening : string;
begin
  Result:=dgettext('units','Error opening %s');
  end;

function rsErrCreating : string;
begin
  Result:=dgettext('units','Error creating %s');
  end;

function rsErrReading : string;
begin
  Result:=dgettext('units','Error reading %s');
  end;

function rsErrWriting : string;
begin
  Result:=dgettext('units','Error writing %s');
  end;

function rsErrClosing : string;
begin
  Result:=dgettext('units','Error closing %s');
  end;

function rsErrSetAttr : string;
begin
  Result:=dgettext('units','Error setting attributes %s');
  end;

function rsErrTimeStamp : string;begin
  Result:=dgettext('units','Error setting timestamp %s');
  end;

function rsErrDirEntry : string;
begin
  Result:=dgettext('units','Error writing directory entry to %s');
  end;

function rsErrEndRecord : string;
begin
  Result:=dgettext('units','Error writing end record to %s');
  end;

function rsErrVerify : string;
begin
  Result:=dgettext('units','Error verifying %s');
  end;

function rsErrCopy : string;
begin
  Result:=dgettext('units','Error copying file to'+sLineBreak+'%s');        //'Fehler beim Kopieren von %';
  end;

function GetCopyErrMsg (AError : integer) : string;
var
  s : string;
begin
  case AError and errAllTypes of
  errError   : s:=dgettext('units','Error: ');
  errWarning : s:=dgettext('units','Warning: ');
  else s:=dgettext('units','Info: ');
    end;
  case AError and errAllCodes of
  errFileCreate : s:=s+dgettext('units','Create file failed');
  errFileOpen   : s:=s+dgettext('units','Open file failed');
  errFileClose  : s:=s+dgettext('units','Close file failed');
  errFileRead   : s:=s+dgettext('units','Read file failed');
  errFileWrite  : s:=s+dgettext('units','Write file failed');
  errFileAttr   : s:=s+dgettext('units','Set file attr. failed');
  errFileFull   : s:=s+dgettext('units','Low disk space');
  errFileGZip   : s:=s+dgettext('units','Illegal GZip header');
  errFileCheck  : s:=s+dgettext('units','Corrupt file');
  errFileECrypt : s:=s+dgettext('units','Encryption failed');
  errFileDCrypt : s:=s+dgettext('units','Decryption failed');
  errFileVerify : s:=s+dgettext('units','Verify failed');
  errLongPath   : s:=s+dgettext('units','Path too long');
  errNotFound   : s:=s+dgettext('units','File not found');
  errFileTS     : s:=s+dgettext('units','Setting of file timestamp failed');
  errStorage    : s:=s+dgettext('units','Error copying file details');
  errFtpRead    : s:=s+dgettext('units','FTP read failed');
  errFtpWrite   : s:=s+dgettext('units','FTP write failed');
  errFtpConnect : s:=s+dgettext('units','FTP connection failed');
  errFtpBroken  : s:=s+dgettext('units','FTP connection closed by server');
  errFtpDatConn : s:=s+dgettext('units','FTP open data connection failed');
  errSignature  : s:=s+dgettext('units','Illegal Zip signature');
  errExtract    : s:=s+dgettext('units','Extracting file failed');
  errFormat     : s:=s+dgettext('units','Unsupported file format');
  errUserBreak  : s:=s+dgettext('units','Terminated by user');
  else s:=s+Format(dgettext('units','Unknown code (%s)'),[IntToHex(AError,8)]);  // should not happen
    end;
  case AError and errAllSources of
  errCopy       : s:=s+dgettext('units',' (Copy)');
  errGZip       : s:=s+dgettext('units',' (GZip)');
  errGUnzip     : s:=s+dgettext('units',' (GUnzip)');
  errZip        : s:=s+dgettext('units',' (Zip)');
  errEnCrypt    : s:=s+dgettext('units',' (Encrypt)');
  errDeCrypt    : s:=s+dgettext('units',' (Decrypt)');
    end;
  Result:=s;
  end;

end.

