(* Copy, pack and unpack files using file transfer protocol (FTP)
   optional with encryption

   using Indy 10 routines
   ======================

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   Acknowledgements:
     ZLibEx (Vers. 1.2.3) and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Feb. 2006
         1.1 - May 2006
         2 - Jul. 2007  - support proxy
         2.1 - Sep. 2007 - support for servers like TOPS-20, see. RFC959
         2.2 - May 2008 - support of MDTM, UTF8 on listing
         2.3 - Aug. 2008 - support SSL/TLS
               Requires: Indy10-Versions of libeay32.dll, ssleay32.dll
                         SSL certificate (created by OpenSSL)
               Added: file match function supporting wildcards
         2.4 - Apr. 2011 - bug fixes on MDTM and LIST
   *)

unit FtpFileTools;

interface

uses Forms, Windows, Classes, SysUtils, ExtFileTools, ZLibex, AesLib, FtGlobal,
     JrUtils, IdFTP, IdAllFTPListParsers, IdComponent;

const
  defTimeout = 10000;   // 10 s
  defFtpBufferSize = 256*1024;

type
  TFtpProxyParams = record
    Server,
    Username,
    Password       : string;
    Port           : integer;
    Mode           : TIdFtpProxyType;
    end;

  TFtpParams = record
    Host,Username,
    Password,Directory  : string;
    Port,
    SecureMode          : integer;
    Passive             : boolean;
    CaseMode            : TTextChange;
    Proxy               : TFtpProxyParams
    end;

  TGetPassword = procedure(var APassword : string) of object;

  TExtFtp = class (TIdFTP)
  private
    FList  : TStringList;
    CMode  : TTextChange;
    fTLS,
    fUTF8  : boolean;
    fCertPwd : string;
    fGetSSLPassword : TGetPassword;
    DCount : integer;            // count "ChangeDir" in "ForceDir"
  protected
    procedure DoAfterLogin; override;
    procedure SetTLSMode(Value : integer);
    procedure GetCertPassword(var Password : string);
  public
    constructor Create (AOwner : TComponent; const Path,Root,Cert,Key,Pwd : string); overload;
    constructor Create (AOwner : TComponent); overload;
    destructor Destroy; override;
    function SetNameCase(FName : string) : string;
    function ConvertToAnsi (const s : string) : string;
    function ConvertToUtf8 (const s : string) : string;
    function GetFileList : boolean;
    function FtpFileExists (const FName : string) : boolean;
    function DeleteFile (const FName : string) : boolean;
    function RenameFile (const OldName,NewName : string) : boolean;
    function DeleteMatchingFiles (const AMask : string) : integer;
    function GetSize (const FName : string) : int64;
    function CheckSize(const FName : string; ASize : int64) : boolean;
    function GetTimeStamp (const FName : string) : TDateTime;
    function GetFileTime (const FName : string) : TFileTime;
    function CheckTimeStamp (const FName : string; tm : TDateTime; Delta : integer) : integer;
    function SetTimeStamp(const ASourceFile,FName : string) : boolean;
    function GetDirList  (DirList : TStringList) : boolean;
    function ChangeToDir(const DName: string) : boolean;
    function ChangeToNewDir(const DName: string) : boolean;
    function ForceDir(DName: string) : boolean;
    function ForceDirUp : boolean;
    function DeleteDir(DName: string) : boolean;
    function GetCurrentDir : string;
    function DeleteTree(DName: string; var DCount : integer) : boolean;
    function ReConnect(const Dir : string) : boolean;
    function Connected : boolean; override;
    function Flush(ASource: TStream; const ADestFile: string; AAppend: boolean) : integer;
    function Fill(const ASourceFile : string; ADest : TStream) : integer;
    function GzFileInfo (Filename: string; var FileInfo : TGzFileInfo) : integer;
    property FileList : TStringList read FList;
    property CaseMode : TTextChange read CMode write CMode;
    property SecureTransfer : integer write SetTLSMode;
    property OnSSLPassword: TGetPassword read fGetSSLPassword write fGetSSLPassword;
    end;

  // Write file via FTP with opt. encryption
  TWriteFtpThread = class (TGZipThread)
  private
    FFtp      : TExtFtp;
    FCompress : boolean;
    FLastResponse : integer;
    FDestDir : string;
  protected
    function WriteFile : integer;
    procedure Execute; override;
    procedure EndThread (Error : integer); override;
    procedure ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
  public
  // ASourceName  = source file,  ADestName =  dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (AFtp : TExtFTP; const ASourceName,ADestDir,ADestName : string;
                        AEncryption : TEncryption; ACompress,ASuspend : Boolean;
                        ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure CancelThread; override;
    property LastResponse : integer read FLastResponse;
    end;

  // Read file via FTP with opt. encryption
  TReadFtpThread = class (TGUnZipThread)
  private
    FFtp      : TExtFtp;
    FCompress : boolean;
    FLastResponse,
    FTimeStamp : integer;
    FFileSize : int64;
  protected
    function ReadFile : integer;
    procedure Execute; override;
    procedure EndThread (Error : integer); override;
    procedure ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
  public
  // ASourceName  = source file,  ADestName =  dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (AFtp : TExtFTP; const ASourceName,ADestName : string;
                        ATimeStamp : integer; AFileSize : int64;
                        AEncryption : TEncryption; ACompress,ASuspend : Boolean;
                        ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure CancelThread; override;
    property LastResponse : integer read FLastResponse;
    end;

// Bei Verzeichnisangaben Slash ergnzen
function IncludeTrailingSlash (Dir : string) : string;
function IncludeLeadingSlash (Dir : string) : string;

// Den letzten Teilnamen extrahieren (siehe auch: ExtractFileName)
function ExtractLastName(const Delimiters,FileName: string): string;

// Unix-Pfad zusammensetzen
function ExpandUnixPath(const APath,AName : string) : string;

// Messages
function rsFtpConn : string;
function rsTryConnect : string;
function rsConnected : string;
function rsConnectErr : string;
function rsFtpHint : string;
function rsProxyHint : string;

implementation

uses IdFTPList, IdReplyRFC, IdReply, IdFTPCommon, Masks,
  IdSSLOpenSSL, IdIOHandler, IdExplicitTLSClientServerBase,
  DateUtils, WinApiUtils, FileUtils, GnuGetText;

{ ------------------------------------------------------------------- }
// Messages
function rsFtpConn : string;    //'FTP-Verbindung';
begin
  Result:=dgettext('units','FTP connection');
  end;

function rsTryConnect : string; //'Verbindung wird hergestellt';
begin
  Result:=dgettext('units','Trying to connect to FTP server');
  end;

function rsConnected : string; //'Verbindung ist hergestellt';
begin
  Result:=dgettext('units','Connection to FTP server established');
  end;

function rsConnectErr : string; //'Verbindung zu FTP-Server konnte nicht hergestellt werden!';
begin
  Result:=dgettext('units','Connection to FTP server failed');
  end;

//'Die Verbindung zu '+'Server: %s:%u'++sLineBreak+'fr den Benutzer: %s'+sLineBreak+'erfordert ein Passwort!';
// Aufruf: Format(rsFtpHint,[host,port,user])
function rsFtpHint : string;
begin
  Result:=dgettext('units','Connection to FTP server: %s:%u'+sLineBreak+'for user: %s'+sLineBreak+'requires a password!');
  end;

function rsProxyHint : string;
begin
  Result:=dgettext('units','Connection to FTP proxy server: %s:%u'+sLineBreak+'for user: %s'+sLineBreak+'requires a password!');
  end;

{ ------------------------------------------------------------------- }
(* Bei Verzeichnisangaben Slash ergnzen *)
function IncludeTrailingSlash (Dir : string) : string;
begin
  if (length(Dir)=0) or (Dir[length(Dir)]='/') then Result:=Dir
  else Result:=Dir+'/';
  end;

function IncludeLeadingSlash (Dir : string) : string;
begin
  if (length(Dir)=0) or (Dir[1]<>'/') then Result:='/'+Dir
  else Result:=Dir;
  end;

// Den letzten Teilnamen extrahieren (siehe auch: ExtractFileName)
function ExtractLastName(const Delimiters,FileName: string): string;
var
  I: Integer;
begin
  I:=LastDelimiter(Delimiters, FileName);
  Result:=Copy(FileName,I+1,MaxInt);
  end;

// Unix-Pfad zusammensetzen
function ExpandUnixPath(const APath,AName : string) : string;
begin
  if length(APath)>0 then Result:=APath+'/'+AName
  else Result:=AName;
  end;

{ ------------------------------------------------------------------- }
constructor TExtFtp.Create (AOwner : TComponent; const Path,Root,Cert,Key,Pwd : string);
var
  sr,sc,sk : string;
begin
  inherited Create(AOwner);
  UseMLIS:=true;
  Passive:=true; fUTF8:=false; fCertPwd:='';
  FList:=TStringList.Create;
  CMode:=tcNone;
// if openSSL libraries and certificates are available, use SSL for TLS support
  fTLS:=(length(Root)>0) and (length(Cert)>0) and (length(Key)>0);
  if fTLS then begin
    try
      TIdSSLContext.Create.Free;
      fTLS:=true;
    except
      fTLS:=false;
      end;
    if fTLS then begin
      sr:=Path+Root; sc:=Path+Cert; sk:=Path+Key;
      fTLS:=FileExists(sr) and FileExists(sc) and FileExists(sk);
      end;
    end;
  if fTLS then begin
    IOHandler:=TIdSSLIOHandlerSocketOpenSSL.Create(self);
    fCertPwd:=Pwd;
    with (IOHandler as TIdSSLIOHandlerSocketOpenSSL)do begin
      with SSLOptions do begin
        RootCertFile:=sr; CertFile:=sc; KeyFile:=sk;
        end;
      OnGetPassword:=GetCertPassword;
      end;
    end
  else IOHandler:=TIdIOHandler.MakeDefaultIOHandler(self);
  ManagedIOHandler:=True;
  UseTLS:=utNoTLSSupport;
  end;

constructor TExtFtp.Create (AOwner : TComponent);
begin
  Create(AOwner,'','','','','');
  end;

destructor TExtFtp.Destroy;
begin
  FList.Free;
  inherited Destroy;
  end;

procedure TExtFtp.DoAfterLogin;
begin
  fUTF8:=IsExtSupported('UTF8');
  if Assigned(FOnAfterClientLogin) then begin
    OnAfterClientLogin(Self);
    end;
  end;

procedure TExtFtp.SetTLSMode(Value : integer);
begin
  if fTLS and (Value>0) then begin
    if Value=1 then UseTLS:=utUseExplicitTLS
    else UseTLS:=utUseRequireTLS;
    DataPortProtection:=ftpdpsPrivate;
    end
  else begin
    UseTLS:=utNoTLSSupport;
    DataPortProtection:=ftpdpsClear;
    end;
  end;

procedure TExtFtp.GetCertPassword(var Password : string);
begin
  if Assigned(fGetSSLPassword) then fGetSSLPassword(Password)  // prompt for password
  else Password:=fCertPwd;
  end;

function TExtFtp.SetNameCase(FName : string) : string;
begin
  Result:=TextChangeCase(FName,CMode);
  end;

function TExtFtp.ConvertToAnsi (const s : string) : string;
begin
  if fUTF8 then Result:=Utf8ToAnsi(s)
  else Result:=s;
  end;

function TExtFtp.ConvertToUtf8 (const s : string) : string;
begin
  if fUTF8 then Result:=AnsiToUtf8(s)
  else Result:=s;
  end;

function TExtFtp.GetFileList : boolean;
var
  i   : integer;
begin
  Result:=true;
  try
    List (nil);
    FList.Clear;
    with DirectoryListing do begin
      // Dateien
      for i:=0 to Count-1 do with Items[i] do if ItemType<>ditDirectory then
        FList.AddObject(ConvertToAnsi(Filename),Items[i]);
      end;
    except
      Result:=false;
      end;
  end;

function TExtFtp.GetDirList (DirList : TStringList) : boolean;
var
  i   : integer;
begin
  Result:=true;
  try
    List (nil);
    DirList.Clear;
    with DirectoryListing do begin
      // Verzeichnisse
      for i:=0 to Count-1 do with Items[i] do
        if (ItemType=ditDirectory) and NotSpecialDir(Filename) then
          DirList.AddObject(ConvertToAnsi(Filename),Items[i]);
      end;
    except
      Result:=false;
      end;
  end;

function TExtFtp.FtpFileExists (const FName : string) : boolean;
begin
  Result:=false;
  if GetFileList then Result:=FileList.IndexOf(FName)>=0;
  end;

function TExtFtp.DeleteFile (const FName : string) : boolean;
begin
  Result:=true;
  try
    Delete(ConvertToUtf8(SetNameCase(FName)));
  except
    Result:=false;
    end;
  end;

function TExtFtp.RenameFile (const OldName,NewName : string) : boolean;
begin
  Result:=false;
  if FtpFileExists(OldName) then begin
    try
      Rename(OldName,NewName);
      Result:=true;
    except
      end;
    end;
  end;

function TExtFtp.DeleteMatchingFiles (const AMask : string) : integer;
var
  i : integer;
begin
  Result:=0;
  with FileList do for i:=0 to Count-1 do if MatchesMask(Strings[i],AMask) then begin
    DeleteFile(ConvertToUtf8(Strings[i]));
    inc(Result);
    end;
  end;

// erfordert vorheriges GetFileList fr das aktuelle Verzeichnis
function TExtFtp.GetSize (const FName : string) : int64;
var
  n : integer;
begin
  n:=FList.IndexOf(SetNameCase(FName));
  if n>=0 then Result:=(FList.Objects[n] as TIdFTPListItem).Size
  else Result:=0;
  end;

function TExtFtp.CheckSize(const FName : string; ASize : int64) : boolean;
begin
  Result:=GetSize(FName)<>ASize;
  end;

// erfordert vorheriges GetFileList fr das aktuelle Verzeichnis
// FTP liefert immer die lokale Zeit auf dem Server
function TExtFtp.GetTimeStamp (const FName : string) : TDateTime;
var
  n : integer;
begin
  n:=FList.IndexOf(SetNameCase(FName));
  if n>=0 then begin
    Result:=(FList.Objects[n] as TIdFTPListItem).ModifiedDate;
    if not ((FCapabilities.Count>0) and FCanUseMLS) or (Result<1) then begin
      // get local time from server because LIST may return time stamps
      // with only date information (time = 0)
      Result:=FileDate(SetNameCase(FName),true);
      end
    end
  else Result:=0;
  end;

function TExtFtp.GetFileTime (const FName : string) : TFileTime;
begin
  Result:=DateTimeToFileTime(GetTimeStamp(FName));
  end;

// Result = 0 : FName nicht gefunden
//          1 : Zeitstempel von FName + Delta < tm
//          2 : Zeitstempel von FName + Delta >= tm
function TExtFtp.CheckTimeStamp (const FName : string; tm : TDateTime; Delta : integer) : integer;
var
  n : integer;
  md : TDateTime;
begin
  n:=FList.IndexOf(SetNameCase(FName));
  if n>=0 then begin
    md:=(FList.Objects[n] as TIdFTPListItem).ModifiedDate;
    if not ((FCapabilities.Count>0) and FCanUseMLS) or (md<1) then begin
      // get local time from server because LIST may return time stamps
      // with only date information (time = 0)
      md:=FileDate(SetNameCase(FName),true);
      end;
    if md>1 then begin
      md:=IncSecond(md,Delta);
      if md<tm then Result:=1 else Result:=2;
      end
    else Result:=0;
    end
  else Result:=0;
  end;

function TExtFtp.SetTimeStamp(const ASourceFile,FName : string) : boolean;
begin
  Result:=true;
  if (FCapabilities.Count>0) and FCanUseMLS then begin
    try
      SetModTimeGMT(ConvertToUtf8(SetNameCase(FName)),
                    FileTimeToDateTime(GetFileLastWriteTime(ASourceFile)));
    except
      Result:=false
      end;
    end
  else begin
    try
      SetModTime(ConvertToUtf8(SetNameCase(FName)),DateTimeFromFileDate(FileAge(ASourceFile)));
    except
      Result:=false
      end;
    end;
  end;

// ChangeDir with optional Utf8 (for compatibility to Delphi 7) 
function TExtFtp.ChangeToDir(const DName: string) : boolean;
begin
  Result:=false;
  try ChangeDir(ConvertToUtf8(DName)); except exit; end;
  Result:=true;
  end;

// Create subdirectory and change to
// Use FTP-Response 257 for change
function TExtFtp.ChangeToNewDir(const DName: string) : boolean;
var
  s : string;
  n : integer;
begin
  Result:=false;
  try
    MakeDir(ConvertToUtf8(DName));
  except
    // Accept "250" as legal response (Some FTP server need this)
    if LastCmdResult.NumericCode<>250 then exit; // error
    end;
  with LastCmdResult do if NumericCode=257 then begin
    s:=Text.Strings[0];
    n:=LastDelimiter('"',s);
    if n>0 then s:=AnsiDequotedStr(copy(s,1,n),'"') else s:=DName;
    end
  else s:=DName;
  try ChangeDir(ConvertToUtf8(s)); except exit; end;
  Result:=true;
  end;

// Create subdirectory(ies)
// Result = true if DName exists or successfully created
// count relative level changes (DCount) for use in "ForceDirUp"
function TExtFtp.ForceDir(DName: string) : boolean;
var
  ok,root : boolean;
  d  : string;
begin
  Result:=true; root:=false;
  DCount:=0;
  if length(DName)>0 then begin
    DName:=SetNameCase(DName);
    ok:=true;
    try
      ChangeDir(ConvertToUtf8(DName));
      DCount:=CountSubStr('/',DName)+1;
    except
      ok:=false;
      end;
    if not ok then begin
      if DName[1]='/' then begin
        try
          ChangeDir ('/');
          System.Delete(DName,1,1);
          ok:=true; root:=true;
        except
          end;
        end
      else ok:=true;
      while ok and (length(DName)>0) do begin
        d:=ReadNxtStr(DName,'/');
        try
          ChangeDir(ConvertToUtf8(d));
          inc(DCount);
        except
          try
            ChangeToNewDir(d);
            inc(DCount);
          except
            ok:=false;
            end;
          end;
        end;
      if root then DCount:=0; // was abs. path
      Result:=ok;
      end
    end;
  end;

// Go up all subdirectories create by last "ForceDir"
function TExtFtp.ForceDirUp: boolean;
var
  i : integer;
begin
  Result:=false;
  for i:=1 to DCount do begin
    try ChangeDirUp; except exit; end;
    end;
  DCount:=0;
  Result:=true;
  end;

function TExtFtp.GetCurrentDir : string;
begin
  Result:=ConvertToAnsi(RetrieveCurrentDir);
  end;

function TExtFtp.DeleteDir (DName : string) : boolean;
begin
  Result:=true;
  try
    RemoveDir(ConvertToUtf8(SetNameCase(DName)));
  except
    Result:=false;
    end;
  end;

// delete tree including all files
function TExtFtp.DeleteTree (DName : string; var DCount : integer) : boolean;
var
  i  : integer;
  ok : boolean;
  DList : TStringList;
begin
  ok:=true;
  DName:=SetNameCase(DName);
  try
    if length(DName)>0 then begin
      try ChangeDir(ConvertToUtf8(DName)); except; end;
      end;
    List (nil);
    // Delete files
    with DirectoryListing do for i:=0 to Count-1 do with Items[i] do
      if (ItemType<>ditDirectory) then begin
        self.Delete(Filename);
        inc(DCount);
        Application.ProcessMessages;
        end;
  except
    if length(DName)>0 then begin
      try ChangeDir(ConvertToUtf8(DName)); except; end;
      end;
    Result:=false; Exit;
    end;
  DList:=TStringList.Create;
  if GetDirList(DList) then begin
    i:=0;
    with DList do while (i<Count) and ok do begin
      ok:=DeleteTree(Strings[i],DCount);
      inc(i);
      end;
    if length(DName)>0 then begin
      try ChangeDirUp; except; end;
      end;
    if ok then begin
      try ok:=DeleteDir(DName); except; end;
      end;
    end
  else ok:=false;
  DList.Free;
  Result:=ok;
  end;

// Check connection and try to reconnect if connection was cancelled
function TExtFtp.ReConnect(const Dir : string) : boolean;
var
  n : integer;
begin
  try
    RetrieveCurrentDir;  // check if still connected
    ChangeToDir(Dir);
    Result:=true;
    Exit;
  except
    // ClosedGracefully
    try Disconnect except end;
    end;
  n:=0;
  while not Connected and (n<5) do begin
    try
      Connect;
      ChangeToDir(Dir);
    except
      inc(n);  // try 5 times
      Sleep(1000);
      end;
    end;
  Result:=n<5;
  end;

// Replaces original function to check if peer has closed connection
function TExtFtp.Connected : boolean;
begin
  try
    Result:=inherited Connected;
  except
    Result:=false;
    end;
  end;

// Stream schreiben
function TExtFtp.Flush(ASource: TStream; const ADestFile: string; AAppend: boolean) : integer;
begin
  if not Connected then begin
    Result:=errError+errFtpBroken; exit;
    end;
  try
    Put(ASource,ConvertToUtf8(SetNameCase(ADestFile)),AAppend);
    if LastCmdResult.NumericCode<400 then Result:=errOK    // siehe RFC 640
    else Result:=errError+errFtpWrite;
  except
    if not Connected then Result:=errError+errFtpBroken
    else Result:=errError+errFtpDatConn;
    end;
  end;

// Stream lesen
function TExtFtp.Fill(const ASourceFile : string; ADest : TStream) : integer;
begin
  if not Connected then begin
    Result:=errError+errFtpBroken; exit;
    end;
  try
    Get(ConvertToUtf8(SetNameCase(ASourceFile)),ADest,false);
    ADest.Seek(0,soFromBeginning);
    Result:=errOK;
  except
    if not Connected then Result:=errError+errFtpBroken
    else Result:=errError+errFtpRead;
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from gz-file *)
function TExtFtp.GzFileInfo (Filename: string; var FileInfo : TGzFileInfo) : integer;
var
  sGz : TFileStream;
  tzn : string;
begin
  Result:=errError+errFileRead;
  try
    tzn:=TempDirectory+tmpGzipName;
    sGz:=TFileStream.Create(tzn,fmCreate); // write to disk
    if (Fill(Filename,sGz)=errOK) then begin
      sGz.Seek(0,soFromBeginning);
      if GetGzInfo(sGz,FileInfo) then result:=errOK;
      end
  finally
    sGz.Free;
    DeleteFile(tzn);
    end;
  end;

{ TWriteFtpThread ------------------------------------------------------------- }
  // Write file via FTP with opt. encryption
constructor TWriteFtpThread.Create (AFtp : TExtFTP;
                                    const ASourceName,ADestDir,ADestName : string;
                                    AEncryption : TEncryption;
                                    ACompress,ASuspend : Boolean;
                                    ABufSize : integer;
                                    APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FFtp:=AFtp; FCompress:=ACompress; FLastResponse:=0;
  if assigned(FCrypting) then FErrFlag:=errEncrypt
  else if FCompress then FErrFlag:=errGZip
  else FErrFlag:=errCopy;
  FFtp.OnWork:=ShowWork;
  FDestDir:=ADestDir;
  end;

destructor TWriteFtpThread.Destroy;
begin
  FFtp.OnWork:=nil;
  inherited Destroy;
  end;

procedure TWriteFtpThread.ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
begin
  if AWorkMode=wmWrite then DoProgress(acFtpWrite,AWorkCount);
  end;

procedure TWriteFtpThread.CancelThread;
begin
  FUserBreak:=true;
  with FFtp do if Connected then begin
    try Abort; except GetInternalResponse end;
    end;
  end;

procedure TWriteFtpThread.EndThread (Error : integer);
begin
//  if (Error and errAllCodes)=errUserBreak then FFtp.Abort;
  inherited EndThread(Error);
  end;

{ ------------------------------------------------------------------- }
function TWriteFtpThread.WriteFile : integer;
var
  ec            : integer;
  FEnc,tff      : boolean;
  FTime         : TFileTime;
  utime         : cardinal;
  sGZip,sEnc    : TStream;
  fp            : int64;
  tzn,ten       : string;
begin
  ec:=errOK;
  if not ExistingFile (SourceName) then ec:=errError+errFileOpen;
  if (ec=errOK) and (length(DestName)=0) then ec:=errError+errFileCreate;
  FTime:=GetFileLastWriteTime(SourceName);
  if ec=errOK then begin
    try
      sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
    except
      ec:=errError+errFileOpen;
      end;
    end;
  if ec=errOK then begin
    FEnc:=assigned(FCrypting);
    tff:=sSource.Size>UMaxMemSize;
    if FCompress then begin
      DoProgress(acCompress,-sSource.Size);   // Reset progress indicator
      // unix time for gzip
      utime:=round(SecsPerDay*(FileTimeToDateTime(FTime)-25569));
      if tff then begin
        tzn:=TempDirectory+tmpGzipName;
        sGZip:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sGZip:=TMemoryStream.Create;
      ec:=GzWriteHeader(sGZip,ExtractFilename(SourceName),Z_DEFLATE,utime,FEnc);
      fp:=sGZip.Position;
     // compress file
      if ec=errOK then begin
        if FEnc then begin
          if tff then begin
            ten:=TempDirectory+tmpEncName;
            sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
            end
          else sEnc:=TMemoryStream.Create;
          ec:=GzCompress(sSource,sEnc);
          if ec=errOK then begin
          // z.B. wenn "sSource" gelockt ist (siehe LockFile)
            if UTotal<>sSource.Size then ec:=errError+errFileRead
            else begin
              DoProgress(acEncrypt,-sEnc.Size);   // Reset progress indicator
              sEnc.Seek(0,soFromBeginning);
              if (ec=errOK) and not FCrypting.EncryptStream(sEnc,sGZip) then
                ec:=errError+errFileECrypt;
              end;
            end;
          sEnc.Free;
          if tff then DeleteFile(ten);
          FCrc:=0;
          end
        else begin
          ec:=GzCompress(sSource,sGZip);
          // z.B. wenn "sSource" gelockt ist (siehe LockFile)
          if (ec=errOK) and (UTotal<>sSource.Size) then ec:=errError+errFileRead;
          end;
        end;
      if ec=errOK then begin
        CTotal:=sGZip.Position-fp;
        if UTotal>MaxCardinal then UTotal:=MaxCardinal;
        ec:=GzWriteTrailer (sGZip,FCrc,TInt64(UTotal).Lo);
        if ec=errOK then begin
          DoProgress (acFtpWrite,-sGZip.Size);  // Reset
          sGZip.Seek(0,soFromBeginning);
          with FFtp do if ReConnect(FDestDir) then ec:=Flush(sGZip,DestName,false)
          else ec:=errFtpConnect;
          FLastResponse:=FFtp.LastCmdResult.NumericCode;
          if (ec=errOK) and FUserBreak then ec:=errUserBreak;
          end;
        end;
      sGZip.Free;
      if tff then DeleteFile(tzn);
      end
    else begin
      if FEnc then begin
        DoProgress(acEncrypt,-sSource.Size);   // Reset progress indicator
        if tff then begin
          ten:=TempDirectory+tmpEncName;
          sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
          end
        else sEnc:=TMemoryStream.Create;
        if not FCrypting.EncryptStream(sSource,sEnc) then ec:=errFileECrypt;
        if ec=errOK then begin
          DoProgress (acFtpWrite,-sEnc.Size);  // Reset
          sEnc.Seek(0,soFromBeginning);
          with FFtp do if ReConnect(FDestDir) then ec:=Flush(sEnc,DestName,false)
          else ec:=errFtpConnect;
          FLastResponse:=FFtp.LastCmdResult.NumericCode;
          end;
        if (ec=errOK) and FUserBreak then ec:=errUserBreak;
        sEnc.Free;
        if tff then DeleteFile(ten);
        end
      else begin
        DoProgress(acFtpWrite,-sSource.Size);   // Reset progress indicator
        ec:=FFtp.Flush(sSource,DestName,false);
        FLastResponse:=FFtp.LastCmdResult.NumericCode;
        if FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        end;
      end;
    try sSource.Free; except end;
    end;
  Result:=ec;
  end;

procedure TWriteFtpThread.Execute;
begin
  EndThread (WriteFile);
  end;

{ TReadFtpThread ------------------------------------------------------------- }
  // Read file via FTP with opt. encryption
constructor TReadFtpThread.Create (AFtp : TExtFTP;
                                   const ASourceName,ADestName : string;
                                   ATimeStamp : integer;
                                   AFileSize : int64;
                                   AEncryption : TEncryption;
                                   ACompress,ASuspend : Boolean;
                                   ABufSize : integer;
                                   APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FFtp:=AFtp; FCompress:=ACompress; FLastResponse:=0;
  FTimeStamp:=ATimeStamp; FFileSize:=AFileSize;
  if assigned(FCrypting) then FErrFlag:=errEncrypt
  else if FCompress then FErrFlag:=errGZip
  else FErrFlag:=errCopy;
  FFtp.OnWork:=ShowWork;
  end;

destructor TReadFtpThread.Destroy;
begin
  FFtp.OnWork:=nil;
  inherited Destroy;
  end;

procedure TReadFtpThread.ShowWork(Sender: TObject; AWorkMode: TWorkMode; AWorkCount: Int64);
begin
  if AWorkMode=wmRead then DoProgress(acFtpRead,AWorkCount);
  end;

procedure TReadFtpThread.CancelThread;
begin
  FUserBreak:=true;
  with FFtp do if Connected then begin
    try Abort; except GetInternalResponse end;
    end;
  end;

procedure TReadFtpThread.EndThread (Error : integer);
begin
//  if (Error and errAllCodes)=errUserBreak then FFtp.Abort;
  inherited EndThread(Error);
  end;

{ ------------------------------------------------------------------- }
function TReadFtpThread.ReadFile : integer;
var
  ec            : integer;
  FEnc,tff      : boolean;
  utime         : cardinal;
  sGZip,sEnc    : TStream;
  dt            : TDateTime;
  fp,n          : int64;
  tzn,ten,s     : string;
begin
  ec:=errOK;
  FEnc:=assigned(FCrypting);
  if (length(DestName)=0) then ec:=errError+errFileCreate;
  if ec=errOK then begin
    try
      sDest:=TFileStream.Create(DestName,fmCreate);
    except
      ec:=errError+errFileCreate;
      end;
    end;
  if ec=errOK then begin
    if FCompress then begin
      tff:=FFileSize>CMaxMemSize;
      DoProgress(acFtpRead,-FFileSize);   // Reset progress indicator
      if tff then begin
        tzn:=TempDirectory+tmpGzipName;
        sGZip:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sGZip:=TMemoryStream.Create;
      ec:=FFtp.Fill(SourceName,sGZip);  // read
      FLastResponse:=FFtp.LastCmdResult.NumericCode;
      if ec=errOK then begin
        sGZip.Seek(0,soFromBeginning);
        // uncompress file
        ec:=GzReadHeader(sGZip,s,uTime,FEnc);
        end;
      if ec=errOK then begin
        fp:=sGZip.Position;
        if FEnc then begin
          DoProgress(acDecrypt,-sGZip.Size);
          if tff then begin
            ten:=TempDirectory+tmpEncName;
            sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
            end
          else sEnc:=TMemoryStream.Create;
          with sGZip do n:=Size-Position-8;
          if FCrypting.DecryptStream(sGZip,sEnc,n) then begin
            DoProgress(acUnCompress,-sEnc.Size);
            sEnc.Seek(0,soFromBeginning);
            ec:=GzUncompress(sEnc,sDest,sEnc.size);
            end
          else ec:=errError+errFileDCrypt;
          sEnc.Free;
          if tff then DeleteFile(ten);
          // Ignore trailer
          end
        else begin
          DoProgress(acUnCompress,-sGZip.Size);
          ec:=GzUnCompress(sGZip,sDest);
          if (ec=errOK) then begin
            CTotal:=sGZip.Position-fp;
            ec:=GzReadTrailer (sGZip);
            end;
          end;
        end;
      sGZip.Free;
      if tff then DeleteFile(tzn);
      end
    else begin
      tff:=FFileSize>UMaxMemSize;
      if FEnc then begin
        DoProgress(acFtpRead,-FFileSize);   // Reset progress indicator
        if tff then begin
          ten:=TempDirectory+tmpEncName;
          sEnc:=TFileStream.Create(ten,fmCreate); // write to disk
          end
        else sEnc:=TMemoryStream.Create;
        ec:=FFtp.Fill(SourceName,sEnc);  // read
        FLastResponse:=FFtp.LastCmdResult.NumericCode;
        if ec=errOK then begin
          DoProgress (acDecrypt,-sEnc.Size);  // Reset
          sEnc.Seek(0,soFromBeginning);
          if not FCrypting.DecryptStream(sEnc,sDest,sEnc.Size) then ec:=errFileDCrypt;
          end;
        if (ec=errOK) and FUserBreak then ec:=errUserBreak;
        sEnc.Free;
        if tff then DeleteFile(ten);
        end
      else begin
        DoProgress(acFtpRead,-FFileSize);   // Reset progress indicator
        ec:=FFtp.Fill(SourceName,sDest);
        FLastResponse:=FFtp.LastCmdResult.NumericCode;
        if FUserBreak then begin
          Sleep(1000);     // auf Abort warten
          ec:=errUserBreak;
          end;
        end;
      end;
    try sDest.Free; except end;
    // Zeitstempel
    if FTimeStamp>0 then begin
    try
      dt:=FileDateToDateTime(FTimeStamp);
    except
      dt:=Now;
      end;
    SetFileLastwriteTime(DestName,DateTimeToFileTime(dt),false);
    end;
    end;
  Result:=ec;
  end;

procedure TReadFtpThread.Execute;
begin
  EndThread (ReadFile);
  end;


end.
